var mongoose = require('mongoose');
var userSchema = mongoose.Schema({
	username: {
		type: String,
		unique: true
	},
	name: {
		firstname: String,
		lastname: String
	},
	regChallenge: String, 
	regTimeStamp: Number,
	publicKey: String,
	loginTimeStamp: Number,
	loginChallenge: String
});
var sessionSchema = mongoose.Schema({
	sessionId: {
		type: String,
		unique: true
	},
	username: String,
	timeStamp: Number
});

var User;
var Session;
var connectionString = 'mongodb://127.0.0.1/Demo'; 
mongoose.connect(connectionString, function(err) {  
	if (err) throw new Error(err);
	console.log("Successfully connected to: " + connectionString);
	User = mongoose.model("User", userSchema);
	Session = mongoose.model("Session", sessionSchema);
});

var express = require('express');
var handlebars = require('express-handlebars').create({defaultLayout: 'main-layout'});
var bodyParser = require('body-parser');
var cookieParser = require('cookie-parser');
var fs = require('fs'); 
var app = express();
app.engine("handlebars", handlebars.engine);
app.set("view engine", "handlebars");

var tlsPrivateKey = fs.readFileSync(__dirname + "/key.pem");
var tlsCertificate = fs.readFileSync(__dirname + "/cert.pem");
var https = require('https');
var options = {
	key: tlsPrivateKey,
	cert: tlsCertificate
}
https.createServer(options, app).listen(443);
console.log("listening on port 443");

var http = require('http');
http.createServer(app).listen(80);
console.log("listening on port 80");

app.use(function(req,res,next) {
	if (!req.secure) {
		res.redirect(301,'https://' + req.headers.host + req.url);
	}
	else {
		next();
	}
});

app.use(function(req,res,next) {
	if (mongoose.connection.readyState != 1) {
		res.status(503).send('SERVER BUSY, TRY AGAIN LATER');
	}
	else {
		next();
	}
});

app.use(express.static('static'));
require(__dirname + '/static/pjcl.js'); // used in client and server code
require(__dirname + '/static/domain-parameters.js'); // used in client and server code

const regRedirTimeout = 30000; // 30 seconds
const loginRedirTimeout = 30000; // 30 seconds
const sessionTimeout = 3600000; // 1 hour

var rbgStateObject = new Object();
var rbgStateInitialized = false;
const rbgSecurityStrength = 128;

const garbageCollectPeriod = 301009; // a little over 5 minutes
const reseedPeriod = 604093; // a little over 10 minutes

function getDevRandomBits(bitLength, f) {
	var byteLength = bitLength / 8;
	var buf = new Buffer(byteLength); 
	(function fillBuf(bufPos) {
		var remaining = byteLength - bufPos;
		if (remaining == 0) {
			f(pjclHex2BitArray(buf.toString('hex')));
			return;
		}
		fs.open('/dev/random', 'r', function(err, fd) {
			if (err) throw new Error(err);
			fs.read(fd, buf, bufPos, remaining, 0, function(err, bytesRead) {
				if (err) throw new Error(err);
				bufPos += bytesRead;
				fs.close(fd, function(err) {
					if (err) throw new Error(err);
					fillBuf(bufPos);
				});
			});
		});
	})(0);
}

getDevRandomBits(rbgSecurityStrength, function(bitArray) {
	pjclRBG128Instantiate(rbgStateObject, bitArray);
	rbgStateInitialized = true;			
	reseedPeriodically(reseedPeriod);
});

function reseedPeriodically(period) {
	setTimeout(getDevRandomBits, period, rbgSecurityStrength, function(bitArray) {
		pjclRBG128Reseed(rbgStateObject, bitArray);
		reseedPeriodically(period);
	});
}

function garbageCollect() {
	var margin = 60000; // 1 minute; may not be necessary
	var cutoffTime = (new Date()).getTime() - regRedirTimeout - margin;
	User.remove({regTimeStamp: {$lt: cutoffTime}, publicKey: {$exists: false}}, function(err) {
		if (err) throw new Error(err);
	});
}

setInterval(garbageCollect, garbageCollectPeriod);

app.use(function(req,res,next) {
	if (!rbgStateInitialized) {
		res.status(503).send('SERVER BUSY, TRY AGAIN LATER');
	}
	else {
		next();
	}
});

app.get('/registration-failure.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Registration failure"
	});
});

app.get('/username-taken.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Username taken"
	});
});

app.get('/username-not-found.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Username not found"
	});
});

app.get('/authentication-failure.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Authentication failure"
	});
});

app.get('/invalid-credential.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Invalid cryptographic credential"
	});
});

app.get('/credential-not-found.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Cryptographic credential not found"
	});
});

app.get('/register.html',function(req,res) {
	res.render("register.handlebars", {});
});

app.get('/',function(req,res) {
	res.redirect(303, "/public-page-1.html");
});

app.get('/please-log-in.html',function(req,res) {
	var destination = req.query.destination;
	if (destination.search(/^[-A-Za-z0-9]+$/) == -1) {
		res.status(404).send('NOT FOUND');
		return;
	}
	res.render("please-log-in.handlebars", {
		destination: destination
	});
});

app.use(bodyParser.urlencoded({ extended: false }));

app.post('/register-username',function(req,res) {
	var username = req.body.username;
	var firstname = req.body.firstname;
	var lastname = req.body.lastname;
	if (
		username.search(/^[A-Za-z0-9]+$/) == -1 ||
		firstname.search(/^[A-Za-z]+$/) == -1 ||
		lastname.search(/^[A-Za-z]+$/) == -1
	) {
		res.redirect(303, "/registration-failure.html");
		return;
	}
	var entropyHex = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));
	var challengeHex = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));

	var user = new User({
		username: username,
		name: {
			firstname: firstname,
			lastname: lastname
		},
		regChallenge: challengeHex, 
		regTimeStamp: (new Date()).getTime()
	});
	user.save(function(err) {
		if (err) {
			res.redirect(303, "/username-taken.html");
		}
		else {
			res.render("reg-redir.handlebars", {
				entropyHex: entropyHex, 
				challengeHex: challengeHex,
				username: username
			});
		}
	});
});

app.post('/register-public-key',function(req,res) {	
	var username = req.body.username;
	var pubKeyHex = req.body.publicKey;
	var sigRHex = req.body.signatureR;
	var sigSHex = req.body.signatureS;
	if (
		username.search(/^[A-Za-z0-9]+$/) == -1 ||
		pubKeyHex.search(/^[A-Za-z0-9]+$/) == -1 ||
		sigRHex.search(/^[A-Fa-f0-9]+$/) == -1 ||
		sigSHex.search(/^[A-Fa-f0-9]+$/) == -1
	) {
		res.redirect(303, "/registration-failure.html");
		return;
	}
	User.findOne({username: username}).select('regTimeStamp regChallenge').exec(function(err, user) {
		if (err) throw new Error(err);
		if (!user) {
			res.redirect(303, "/registration-failure.html");
			return;
		}
		var now = (new Date()).getTime();
		if (!user.regTimeStamp || (now - user.regTimeStamp > regRedirTimeout)) { 
			res.redirect(303, "/registration-failure.html");
			return;
		}
		var challengeHex = user.regChallenge;
		if (!challengeHex) {
			res.redirect(303, "/registration-failure.html");
			return;
		}
		var challengeBitArray = pjclHex2BitArray(challengeHex);
		var p = pjclHex2BigInt(pHex);
		var q = pjclHex2BigInt(qHex);
		var g = pjclHex2BigInt(gHex);
		var y = pjclHex2BigInt(pubKeyHex);
		if (!pjclFFCValidatePublicKey(p,q,g,y)) {
			res.redirect(303, "/registration-failure.html");
			return;
		}
		var r = pjclHex2BigInt(sigRHex);
		var s = pjclHex2BigInt(sigSHex);
		if (!pjclDSAVerifyMsg(p,q,g,y,challengeBitArray,r,s)) {
			res.redirect(303, "/registration-failure.html");
			return;
		}
		user.publicKey = pubKeyHex;
		user.regChallenge = "";
		user.save(function (err) {
			if (err) throw new Error(err);
			var sessionId = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));
			var session = new Session({
				sessionId: sessionId, 
				username: username,
				timeStamp: new Date().getTime()
			});
			session.save(function (err) {
				if (err) throw new Error(err);
				res.cookie('session', sessionId, {httpOnly: true, secure: true});
				res.redirect(303, "/private-page-1.html");
			});
		});
	});
});

app.post('/check-username',function(req,res) {
	var username = req.body.username;
	var destination = req.body.destination || "private-page-1";
	if (destination.search(/^[-A-Za-z0-9]+$/) == -1) {
		res.redirect(303, "/authentication-failure.html");
		return;
	}
	if (username.search(/^[A-Za-z0-9]+$/) == -1) {
		res.redirect(303, "/username-not-found.html");
		return;
	}
	User.findOne({username: username}).exec(function(err, user) {
		if (err) throw new Error(err);
		if (!user) {
			res.redirect(303, "/username-not-found.html");
			return;
		}
		var entropyHex = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));
		var challengeHex = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));
		user.loginChallenge = challengeHex;
		user.loginTimeStamp = new Date().getTime();
		user.save(function(err) {
			if (err) throw new Error(err);
			res.render("login-redir.handlebars", {
				entropyHex: entropyHex, 
				challengeHex: challengeHex,
				username: username,
				destination: destination
			});
		});
	});
});

app.post('/verify-signature',function(req,res) { 
	var username = req.body.username;
	var pubKeyHex = req.body.publicKey;
	var sigRHex = req.body.signatureR;
	var sigSHex = req.body.signatureS;
	var destination = req.body.destination;
	if (
		username.search(/^[A-Za-z0-9]+$/) == -1 ||
		pubKeyHex.search(/^[A-Fa-f0-9]+$/) == -1 ||
		sigRHex.search(/^[A-Fa-f0-9]+$/) == -1 ||
		sigSHex.search(/^[A-Fa-f0-9]+$/) == -1 ||
		destination.search(/^[-A-Za-z0-9]+$/) == -1
	) {
		res.redirect(303, "/authentication-failure.html");
		return;
	}
	User.findOne({username: username}).select('loginTimeStamp loginChallenge publicKey').exec(function(err, user) {
		if (err) throw new Error(err);
		if (!user) {
			res.redirect(303, "/authentication-failure.html");
			return;
		}
		var now = (new Date()).getTime();
		if (!user.loginTimeStamp || (now - user.loginTimeStamp > loginRedirTimeout)) { 
				res.redirect(303, "/authentication-failure.html");
				return;
			}
		var challengeHex = user.loginChallenge;
		if (!challengeHex) {
			res.redirect(303, "/authentication-failure.html");
			return;
		}
		if (pubKeyHex != user.publicKey) { 
			res.redirect(303, "/authentication-failure.html");
			return;
		}
		var p = pjclHex2BigInt(pHex);
		var q = pjclHex2BigInt(qHex);
		var g = pjclHex2BigInt(gHex);
		var y = pjclHex2BigInt(pubKeyHex);
		var msg = pjclHex2BitArray(user.loginChallenge);
		var r = pjclHex2BigInt(sigRHex);
		var s = pjclHex2BigInt(sigSHex);
		if (!pjclDSAVerifyMsg(p,q,g,y,msg,r,s)) {  
			res.redirect(303, "/invalid-credential.html");
			return;
		}
		user.loginChallenge = "";
		user.save(function (err) {
			if (err) throw new Error(err);
			var sessionId = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));
			var session = new Session({
				sessionId: sessionId,
				username: username,
				timeStamp: new Date().getTime()
			});
			session.save(function (err) {
				if (err) throw new Error(err);
				res.cookie('session', sessionId, {httpOnly: true, secure: true});
				res.redirect(303, `/${destination}.html`);
			});
		});
	});
});

app.use(cookieParser());

app.get('/logout',function(req,res) {
	var sessionId = req.cookies.session;
	if (sessionId) {
		res.clearCookie('session');
		Session.findOne({sessionId: sessionId}).exec(function(err, session) {
			if (err) throw new Error(err);
			if (session) {
				session.remove(function (err) {  
					if (err) throw new Error(err);
					res.redirect(303, "/");
				});
			}
			else {
				res.redirect(303, "/");
			}
		});
	}
});

function checkIfLoggedIn(req,res,next) {
	var sessionId = req.cookies.session;
	if (
		!sessionId ||
		sessionId.search(/^[A-Fa-f0-9]+$/) == -1
	) {
		res.locals.loggedIn = false;
		next();
		return;
	}   
	Session.findOne({sessionId: sessionId}).select('timeStamp username').exec(function(err, session) {
		if (err) throw new Error(err);
		if (!session) {
			res.locals.loggedIn = false;
			next();
			return;
		}
		var now = (new Date()).getTime();
		if (!session.timeStamp || (now - session.timeStamp > sessionTimeout)) { 
			res.locals.loggedIn = false;
			next();
			return;
		}
		User.findOne({username: session.username}).select('name').exec(function(err, user) {
			if (err) throw new Error(err);
			if (!user) {
				res.locals.loggedIn = false;
				next();
				return;
			}
			res.locals.loggedIn = true;
			res.locals.fullName = `${user.name.firstname} ${user.name.lastname}`;
			next();
		});
	});
};

app.use(checkIfLoggedIn);

var publicPageNames = [
	"public-page-1",
	"public-page-2",
	"public-page-3"
];
publicPageNames.forEach(function(pageName) {
	app.get(`/${pageName}.html`,function(req,res) {
		res.render(`${pageName}.handlebars`);
	});
});

var privatePageNames = [
	"private-page-1",
	"private-page-2",
	"private-page-3"
];
privatePageNames.forEach(function(pageName) {
	app.get(`/${pageName}.html`,function(req,res) {
		if (res.locals.loggedIn) {
			res.render(`${pageName}.handlebars`);
		}
		else {
			res.redirect(303,`/please-log-in.html?destination=${pageName}`);
		}
	});
});

app.use(function(req,res) {
	res.status(404).send('NOT FOUND');
});

app.use(function(err,req,res,next) {
	console.log("Error: " + err.stack);
	res.status(500).send('INTERNAL ERROR');
});
