var userTable = new Object();
var sessionTable = new Object();

var express = require('express');
var handlebars = require('express-handlebars').create({defaultLayout: 'main-layout'});
var bodyParser = require('body-parser');
var cookieParser = require('cookie-parser');
var fs = require('fs');
var app = express();
app.engine("handlebars", handlebars.engine);
app.set("view engine", "handlebars");

var tlsPrivateKey = fs.readFileSync(__dirname + "/key.pem");
var tlsCertificate = fs.readFileSync(__dirname + "/cert.pem");
var https = require('https');
var options = {
	key: tlsPrivateKey,
	cert: tlsCertificate
}
https.createServer(options, app).listen(443);
console.log("listening on port 443");

var http = require('http');
http.createServer(app).listen(80);
console.log("listening on port 80");

app.use(function(req,res,next) {
	if (!req.secure) {
		res.redirect(301,'https://' + req.headers.host + req.url);
	}
	else {
		next();
	}
});

app.use(express.static('static'));
require(__dirname + '/static/pjcl.js'); // used in client and server code
require(__dirname + '/static/domain-parameters.js'); // used in client and server code

const regRedirTimeout = 30000; // 30 seconds
const loginRedirTimeout = 30000; // 30 seconds
const sessionTimeout = 3600000; // 1 hour

var rbgStateObject = new Object();
var rbgStateInitialized = false;
const rbgSecurityStrength = 128;

const garbageCollectPeriod = 301009; // a little over 5 minutes
const reseedPeriod = 604093; // a little over 10 minutes

function getDevRandomBits(bitLength, f) {
	var byteLength = bitLength / 8;
	var buf = new Buffer(byteLength); 
	(function fillBuf(bufPos) {
		var remaining = byteLength - bufPos;
		if (remaining == 0) {
			f(pjclHex2BitArray(buf.toString('hex')));
			return;
		}
		fs.open('/dev/random', 'r', function(err, fd) {
			if (err) throw new Error(err);
			fs.read(fd, buf, bufPos, remaining, 0, function(err, bytesRead) {
				if (err) throw new Error(err);
				bufPos += bytesRead;
				fs.close(fd, function(err) {
					if (err) throw new Error(err);
					fillBuf(bufPos);
				});
			});
		});
	})(0);
}

getDevRandomBits(rbgSecurityStrength, function(bitArray) {
	pjclRBG128Instantiate(rbgStateObject, bitArray);
	rbgStateInitialized = true;			
	reseedPeriodically(reseedPeriod);
});

function reseedPeriodically(period) {
	setTimeout(getDevRandomBits, period, rbgSecurityStrength, function(bitArray) {
		pjclRBG128Reseed(rbgStateObject, bitArray);
		reseedPeriodically(period);
	});
}

function garbageCollect() {
	var margin = 60000; // 1 minute; may not be necessary
	var cutoffTime = (new Date()).getTime() - regRedirTimeout - margin;
	for (var username in userTable) {
		var userRecord = userTable[username];
		if (userRecord.publicKey === undefined && userRecord.regTimeStamp < cutoffTime) {
			delete userTable[username];
		}
	}
}

setInterval(garbageCollect, garbageCollectPeriod);

app.use(function(req,res,next) {
	if (!rbgStateInitialized) {
		res.status(503).send('SERVER BUSY, TRY AGAIN LATER');
	}
	else {
		next();
	}
});

app.get('/registration-failure.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Registration failure"
	});
});

app.get('/username-taken.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Username taken"
	});
});

app.get('/username-not-found.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Username not found"
	});
});

app.get('/authentication-failure.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Authentication failure"
	});
});

app.get('/invalid-credential.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Invalid cryptographic credential"
	});
});

app.get('/credential-not-found.html',function(req,res) {
	res.render("error.handlebars", {
		msg: "Cryptographic credential not found"
	});
});

app.get('/register.html',function(req,res) {
	res.render("register.handlebars", {});
});

app.get('/',function(req,res) {
	res.redirect(303, "/public-page-1.html");
});

app.get('/please-log-in.html',function(req,res) {
	var destination = req.query.destination;
	if (destination.search(/^[-A-Za-z0-9]+$/) == -1) {
		res.status(404).send('NOT FOUND');
		return;
	}
	res.render("please-log-in.handlebars", {
		destination: destination
	});
});

app.use(bodyParser.urlencoded({ extended: false }));

app.post('/register-username',function(req,res) {
	var username = req.body.username;
	var firstname = req.body.firstname;
	var lastname = req.body.lastname;
	if (
		username.search(/^[A-Za-z0-9]+$/) == -1 ||
		firstname.search(/^[A-Za-z]+$/) == -1 ||
		lastname.search(/^[A-Za-z]+$/) == -1
	) {
		res.redirect(303, "/registration-failure.html");
		return;
	}
	if (userTable[username]) {
		res.redirect(303, "/username-taken.html");
		return;
	}
	var entropyHex = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));
	var challengeHex = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));
	userTable[username] = {
		firstname: firstname, 
		lastname: lastname, 
		regChallenge: challengeHex, 
		regTimeStamp: (new Date()).getTime()
	};
	res.render("reg-redir.handlebars", {
		entropyHex: entropyHex, 
		challengeHex: challengeHex,
		username: username
	});
});

app.post('/register-public-key',function(req,res) {
	var username = req.body.username;
	var pubKeyHex = req.body.publicKey;
	var sigRHex = req.body.signatureR;
	var sigSHex = req.body.signatureS;
	if (
		username.search(/^[A-Za-z0-9]+$/) == -1 ||
		pubKeyHex.search(/^[A-Fa-f0-9]+$/) == -1 ||
		sigRHex.search(/^[A-Fa-f0-9]+$/) == -1 ||
		sigSHex.search(/^[A-Fa-f0-9]+$/) == -1
	) {
		res.redirect(303, "/registration-failure.html");
		return;
	}
	var userRecord = userTable[username];
	if (!userRecord) {
		res.redirect(303, "/registration-failure.html");
		return;
	}
	var now = (new Date()).getTime();
	if (!userRecord.regTimeStamp || (now - userRecord.regTimeStamp > regRedirTimeout)) { 
		res.redirect(303, "/registration-failure.html");
		return;
	}
	var challengeHex = userRecord.regChallenge;
	if (!challengeHex) {
		res.redirect(303, "/registration-failure.html");
		return;
	}
	var challengeBitArray = pjclHex2BitArray(challengeHex);
	var p = pjclHex2BigInt(pHex);
	var q = pjclHex2BigInt(qHex);
	var g = pjclHex2BigInt(gHex);
	var y = pjclHex2BigInt(pubKeyHex);
	if (!pjclFFCValidatePublicKey(p,q,g,y)) {
		res.redirect(303, "/registration-failure.html");
		return;
	}
	var r = pjclHex2BigInt(sigRHex);
	var s = pjclHex2BigInt(sigSHex);
	if (!pjclDSAVerifyMsg(p,q,g,y,challengeBitArray,r,s)) {
		res.redirect(303, "/registration-failure.html");
		return;
	}
	userRecord.publicKey = pubKeyHex;
	delete userRecord.regChallenge;	 
	var sessionId = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));
	var sessionRecord = new Object();
	sessionRecord.username = username;
	sessionRecord.timeStamp = (new Date()).getTime();
	sessionTable[sessionId] = sessionRecord;
	res.cookie('session', sessionId, {httpOnly: true, secure: true});
	res.redirect(303, "/private-page-1.html");
});

app.post('/check-username',function(req,res) {
	var username = req.body.username;
	var destination = req.body.destination || "private-page-1";
	if (destination.search(/^[-A-Za-z0-9]+$/) == -1) {
		res.redirect(303, "/authentication-failure.html");
		return;
	}
	if (username.search(/^[A-Za-z0-9]+$/) == -1) {
		res.redirect(303, "/username-not-found.html");
		return;
	}
	if (!userTable[username]) {
		res.redirect(303, "/username-not-found.html");
		return;
	}
	var entropyHex = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));
	var challengeHex = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));
	var userRecord = userTable[username];
	userRecord.loginChallenge = challengeHex;
	userRecord.loginTimeStamp = (new Date()).getTime();
	res.render("login-redir.handlebars", {
		entropyHex: entropyHex, 
		challengeHex: challengeHex,
		username: username,
		destination: destination
	});
});

app.post('/verify-signature',function(req,res) {
	var username = req.body.username;
	var pubKeyHex = req.body.publicKey;
	var sigRHex = req.body.signatureR;
	var sigSHex = req.body.signatureS;
	var destination = req.body.destination;
	if (
		username.search(/^[A-Za-z0-9]+$/) == -1 ||
		pubKeyHex.search(/^[A-Fa-f0-9]+$/) == -1 ||
		sigRHex.search(/^[A-Fa-f0-9]+$/) == -1 ||
		sigSHex.search(/^[A-Fa-f0-9]+$/) == -1 ||
	    destination.search(/^[-A-Za-z0-9]+$/) == -1
	) {
		res.redirect(303, "/authentication-failure.html");
		return;
	}
	var userRecord = userTable[username];
	if (!userRecord) {
		res.redirect(303, "/authentication-failure.html");
		return;
	}
	var now = (new Date()).getTime();
	if (!userRecord.loginTimeStamp || (now - userRecord.loginTimeStamp > loginRedirTimeout)) { 
		res.redirect(303, "/authentication-failure.html");
		return;
	}
	var challengeHex = userRecord.loginChallenge;
	if (!challengeHex) {
		res.redirect(303, "/authentication-failure.html");
		return;
	}
	if (pubKeyHex != userRecord.publicKey) {
		res.redirect(303, "/authentication-failure.html");
		return;
	}
	var p = pjclHex2BigInt(pHex);
	var q = pjclHex2BigInt(qHex);
	var g = pjclHex2BigInt(gHex);
	var y = pjclHex2BigInt(pubKeyHex);
	var msg = pjclHex2BitArray(challengeHex);
	var r = pjclHex2BigInt(sigRHex);
	var s = pjclHex2BigInt(sigSHex);
	if (!pjclDSAVerifyMsg(p,q,g,y,msg,r,s)) {
		res.redirect(303, "/invalid-credential.html");
		return;
	}
	delete userRecord.loginChallenge;
	var sessionId = pjclBitArray2Hex(pjclRBGGen(rbgStateObject,rbgSecurityStrength,rbgSecurityStrength));
	var sessionRecord = new Object();
	sessionRecord.username = username;
	sessionRecord.timeStamp = (new Date()).getTime();
	sessionTable[sessionId] = sessionRecord;
	res.cookie('session', sessionId, {httpOnly: true, secure: true});
	res.redirect(303, `/${destination}.html`);
});

app.use(cookieParser());

app.get('/logout',function(req,res) {
	var sessionId = req.cookies.session;
	if (sessionId) {
		delete sessionTable[sessionId];
		res.clearCookie('session');
	}
	res.redirect(303, "/");
});

var checkIfLoggedIn = function(req,res,next) {
	var sessionId = req.cookies.session;
	if (
		!sessionId ||
		sessionId.search(/^[A-Fa-f0-9]+$/) == -1
	) {
		res.locals.loggedIn = false;
		next();
		return;
	}
	var sessionRecord = sessionTable[sessionId];
	if (!sessionRecord) {
		res.locals.loggedIn = false;
		next();
		return;
	}
	var now = (new Date()).getTime();
	if (now - sessionRecord.timeStamp > sessionTimeout) {
		res.locals.loggedIn = false;
		next();
		return;
	}
	var username = sessionRecord.username;
	var userRecord = userTable[username];
	if (!userRecord) {
		res.locals.loggedIn = false;
		next();
		return;
	}
	res.locals.loggedIn = true;
	res.locals.fullName = `${userRecord.firstname} ${userRecord.lastname}`;
	next();
}

app.use(checkIfLoggedIn);

var publicPageNames = [
	"public-page-1",
	"public-page-2",
	"public-page-3"
];
publicPageNames.forEach(function(pageName) {
	app.get(`/${pageName}.html`,function(req,res) {
		res.render(`${pageName}.handlebars`);
	});
});

var privatePageNames = [
	"private-page-1",
	"private-page-2",
	"private-page-3"
];
privatePageNames.forEach(function(pageName) {
	app.get(`/${pageName}.html`,function(req,res) {
		if (res.locals.loggedIn) {
			res.render(`${pageName}.handlebars`);
		}
		else {
			res.redirect(303,`/please-log-in.html?destination=${pageName}`);
		}
	});
});

app.use(function(req,res) {
	res.status(404).send('NOT FOUND');
});

app.use(function(err,req,res,next) {
	console.log(err.stack);
	res.status(500).send('INTERNAL ERROR');
});

